r"""

Expectation Propagation
^^^^^^^^^^^^^^^^^^^^^^^

Introduction
------------

This module implements the building-blocks for EP inference: EP parameter
fitting, log of the marginal likelihood, and derivative of the log of the
marginal likelihood.

Private interface
-----------------
"""

from .ep import EP

__all__ = ['EP']



from __future__ import division

from numpy import clip, exp
from numpy_sugar import is_all_finite

from optimix import Function, Scalar, maximize_scalar

from .core import LMMCore


class LMM(LMMCore, Function):
    r"""Fast Linear Mixed Models inference via maximum likelihood.

    It models

    .. math::

        \mathbf y \sim \mathcal N\Big(~ \mathrm X\boldsymbol\beta;~
          s \big(
            (1-\delta)
              \mathrm Q \mathrm S \mathrm Q^{\intercal} +
            \delta \mathrm I
          \big)
        ~\Big)

    for which :math:`\mathrm Q\mathrm S\mathrm Q^{\intercal}=\tilde{\mathrm K}`
    is the eigen decomposition of :math:`\tilde{\mathrm K}`.

    Args:
        y (array_like): outcome.
        X (array_like): covariates as a two-dimensional array.
        QS (tuple): economic eigen decompositon ((Q0, Q1), S0).

    Examples
    --------

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from limix_inference.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.learn(progress=False)
        >>> print('%.3f' % lmm.lml())
        -3.649

    One can also specify which parameters should be fitted:

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from limix_inference.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fix('delta')
        >>> lmm.fix('scale')
        >>> lmm.delta = 0.5
        >>> lmm.scale = 1
        >>> lmm.learn(progress=False)
        >>> print('%.3f' % lmm.lml())
        -4.232
        >>> print('%.1f' % lmm.heritability)
        0.5
        >>> lmm.unfix('delta')
        >>> lmm.learn(progress=False)
        >>> print('%.3f' % lmm.lml())
        -2.838
        >>> print('%.1f' % lmm.heritability)
        0.0
    """

    def __init__(self, y, X, QS):
        LMMCore.__init__(self, y, X, QS)
        Function.__init__(self, logistic=Scalar(0.0))

        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the phenotype.")

        self._fix = dict(beta=False, scale=False)
        self._delta = 0.5
        self._scale = LMMCore.scale.fget(self)  # pylint: disable=E1101
        self.set_nodata()

    def fix(self, var_name):
        if var_name == 'delta':
            super(LMM, self).fix('logistic')
        else:
            self._fix[var_name] = True

    def unfix(self, var_name):
        if var_name == 'delta':
            super(LMM, self).unfix('logistic')
        else:
            self._fix[var_name] = False

    @property
    def scale(self):
        if self._fix['scale']:
            return self._scale
        return LMMCore.scale.fget(self)  # pylint: disable=E1101

    @scale.setter
    def scale(self, v):
        self._scale = v

    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, delta):
        self._delta = delta

    def copy(self):
        # pylint: disable=W0212
        o = LMM.__new__(LMM)

        LMMCore.__init__(o, self._y, self.X, self._QS)
        Function.__init__(o, logistic=Scalar(self.variables().get('logistic').value))

        o._fix = self._fix.copy()
        o._delta = self._delta
        o._scale = self._scale

        o.set_nodata()
        return o

    def _get_delta(self):
        v = clip(self.variables().get('logistic').value, -20, 20)
        x = 1 / (1 + exp(-v))
        return clip(x, 1e-5, 1 - 1e-5)

    @property
    def heritability(self):
        t = (self.fixed_effects_variance + self.genetic_variance +
             self.environmental_variance)
        return self.genetic_variance / t

    @property
    def fixed_effects_variance(self):
        return self.m.var()

    @property
    def genetic_variance(self):
        return self.scale * (1 - self.delta)

    @property
    def environmental_variance(self):
        return self.scale * self.delta

    def learn(self, progress=True):
        maximize_scalar(self, progress=progress)
        self.update()
        self.delta = self._get_delta()

    def value(self):
        self.delta = self._get_delta()
        return self.lml()

    def lml(self):
        self.delta = self._get_delta()
        return LMMCore.lml(self)



from __future__ import division

from numpy.random import RandomState
from numpy_sugar.linalg import sum2diag
from numpy_sugar import epsilon
from numpy_sugar.random import multivariate_normal


class GPSampler(object): # pylint: disable=R0903
    r"""Sample from a Gaussian Process.

    Outcome modelled via

    .. math::

        \mathbf y \sim \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K)
        \mathrm d\mathbf z.

    Args:
        mean (:class:`optimix.Function`): mean function.
                                          (Refer to :doc:`mean`.)
        cov (:class:`optimix.Function`): covariance function.
                                         (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from limix_inference.example import offset_mean
        >>> from limix_inference.example import linear_eye_cov
        >>> from limix_inference.random import GPSampler
        >>>
        >>> random = RandomState(1)
        >>>
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GPSampler(mean, cov).sample(random)
        >>> print(y[:5])
        [ 1.67474605  1.24736152 -0.77509523  1.88952893  1.62847894]
    """
    def __init__(self, mean, cov):
        self._mean = mean
        self._cov = cov

    def sample(self, random_state=None):
        if random_state is None:
            random_state = RandomState()

        m = self._mean.feed('sample').value()
        K = self._cov.feed('sample').value().copy()

        sum2diag(K, +epsilon.small, out=K)

        return multivariate_normal(m, K, random_state)



from __future__ import division

from numpy import append, dot, empty, log, full
from numpy import sum as npsum

from numpy_sugar.linalg import solve

LOG2PI = 1.837877066409345339081937709124758839607238769531250

class FastScanner(object): # pylint: disable=R0903
    r"""Fast inference over multiple covariates.

    Let :math:`\tilde{\mathrm M}_i` be a column-matrix of fixed-effect
    :math:`i`.
    It fits :math:`\alpha_i` and refits :math:`\boldsymbol\beta` and :math:`s`
    for each fixed-effect :math:`i` in order to compute LMLs:

    .. math::

        \mathbf y \sim \mathcal N\big(~ \mathrm M\boldsymbol\beta
        + \tilde{\mathrm M}_i \alpha_i;~ s \mathrm K ~\big)

    Args:

        y (array_like): real-valued outcome.
        M (array_like): matrix of covariates.
        QS (tuple): economic eigen decomposition ``((Q0, Q1), S0)``.
    """

    def __init__(self, y, M, QS, delta):

        self._QS = QS
        self._diags = [(1 - delta) * QS[1] + delta, delta]

        yTQ = [dot(y.T, Q) for Q in QS[0]]
        MTQ = [dot(M.T, Q) for Q in QS[0]]

        self._yTQdiag = [l / r for (l, r) in zip(yTQ, self._diags)]

        self._a = [(i**2 / j).sum() for (i, j) in zip(yTQ, self._diags)]
        self._b = [dot(i, j.T) for(i, j) in zip(self._yTQdiag, MTQ)]

        self._MTQdiag = [i / j for (i, j) in zip(MTQ, self._diags)]

        nc = M.shape[1]

        self._C = [empty((nc + 1, nc + 1)), empty((nc + 1, nc + 1))]

        for i in range(2):
            self._C[i][:-1, :-1] = dot(self._MTQdiag[i], MTQ[i].T)

    def _static_lml(self):
        n = self._QS[0][0].shape[0]
        p = len(self._diags[0])
        static_lml = -n * LOG2PI - n
        static_lml -= npsum(log(self._diags[0]))
        static_lml -= (n - p) * log(self._diags[1])
        return static_lml

    def fast_scan(self, markers):
        r"""LMLs of markers by fitting scale and fixed-effect sizes parameters.

        Args:

            markers (array_like): matrix of fixed-effects across columns.

        Returns:
            tuple: LMLs and effect-sizes, respectively.
        """
        assert markers.ndim == 2

        mTQ = [dot(markers.T, Q) for Q in self._QS[0]]

        bm = [dot(i, j.T) for (i, j) in zip(self._yTQdiag, mTQ)]

        c_01 = [dot(i, j.T) for (i, j) in zip(self._MTQdiag, mTQ)]
        c_11 = [npsum((i / j) * i, axis=1) for (i, j) in zip(mTQ, self._diags)]

        lmls = full(markers.shape[1], self._static_lml())
        effect_sizes = empty(markers.shape[1])

        for i in range(markers.shape[1]):

            b00m = append(self._b[0], bm[0][i])
            b11m = append(self._b[1], bm[1][i])

            self._C[0][:-1, -1] = c_01[0][:, i]
            self._C[1][:-1, -1] = c_01[1][:, i]

            self._C[0][-1, :-1] = self._C[0][:-1, -1]
            self._C[1][-1, :-1] = self._C[1][:-1, -1]

            self._C[0][-1, -1] = c_11[0][i]
            self._C[1][-1, -1] = c_11[1][i]

            beta = solve(self._C[1] - self._C[0], b11m - b00m)
            effect_sizes[i] = beta[-1]

            p0 = self._a[1] - 2 * b11m.dot(beta) + beta.dot(self._C[1].dot(beta))
            p1 = self._a[0] - 2 * b00m.dot(beta) + beta.dot(self._C[0]).dot(beta)

            scale = (p0 + p1) / markers.shape[0]

            lmls[i] -= markers.shape[0] * log(scale)

        lmls /= 2
        return lmls, effect_sizes



"""
**************
Random sampler
**************

Introduction
^^^^^^^^^^^^

Helper classes and functions for sampling: :class:`.GGPSampler`,
:class:`.GPSampler`.

Public interface
^^^^^^^^^^^^^^^^
"""

from .ggp import GGPSampler
from .gp import GPSampler
from .canonical import bernoulli_sample, binomial_sample, poisson_sample

__all__ = ['GGPSampler', 'GPSampler', 'bernoulli_sample', 'binomial_sample',
           'poisson_sample']



from numpy.testing import assert_allclose

from limix_inference.mean import LinearMean
from optimix.testing import Assertion


def test_offsetmean_optimix():
    item0 = [5.1, 1.0]
    item1 = [2.1, -0.2]

    cov = LinearMean(2)

    a = Assertion(lambda: cov, item0, item1, 0.0, effsizes=[0.5, 1.0])
    a.assert_layout()
    a.assert_gradient()

    cov.effsizes = [1.0, -1.0]
    assert_allclose(cov.effsizes, [1.0, -1.0])



from __future__ import division

from numpy import dot, empty, sqrt
from numpy import sum as npsum
from numpy_sugar.linalg import cho_solve, ddot, dotd, sum2diag
from scipy.linalg import cho_factor


class Posterior(object):
    r"""EP posterior.

    It is given by

    .. math::

        \mathbf z \sim \mathcal N\left(\Sigma (\tilde{\mathrm T}
          \tilde{\boldsymbol\mu} + \mathrm K^{-1}\mathbf m),
          (\tilde{\mathrm T} + \mathrm K^{-1})^{-1}\right).
    """
    def __init__(self, site):
        n = len(site.tau)
        self.tau = empty(n)
        self.eta = empty(n)
        self._site = site
        self._mean = None
        self._QS = None

    def set_prior_mean(self, mean):
        self._mean = mean

    def set_prior_cov(self, QS):
        self._QS = QS

    def prior_mean(self):
        return self._mean

    def prior_cov(self):
        return self._QS

    def initialize(self):
        r"""Initialize the mean and covariance of the posterior.

        Given that :math:`\tilde{\mathrm T}` is a matrix of zeros right before
        the first EP iteration, we have

        .. math::

            \boldsymbol\mu = \mathrm K^{-1} \mathbf m ~\text{ and }~
            \Sigma = \mathrm K

        as the initial posterior mean and covariance.
        """
        self.tau[:] = 1 / npsum((self._QS[0] * sqrt(self._QS[1]))**2,
                                axis=1)
        self.eta[:] = self._mean
        self.eta[:] *= self.tau

    def L(self):
        r"""Cholesky decomposition of :math:`\mathrm B`.

        .. math::

            \mathrm B = \mathrm Q^{\intercal}\tilde{\mathrm{T}}\mathrm Q
                + \mathrm{S}^{-1}
        """
        QS = self._QS
        B = dot(QS[0].T, ddot(self._site.tau, QS[0], left=True))
        sum2diag(B, 1. / QS[1], out=B)
        return cho_factor(B, lower=True)[0]

    # def QS(self):
    #     r"""Eigen decomposition of :math:`\mathrm K`"""
    #     QS = economic_qs(self._cov)
    #     return QS[0][0], QS[1]

    def _BiQt(self):
        return cho_solve(self.L(), self._QS[0].T)

    def update(self):
        QS = self._QS
        cov = dot(ddot(QS[0], QS[1], left=False), QS[0].T)

        BiQt = self._BiQt()
        TK = ddot(self._site.tau, cov, left=True)
        BiQtTK = dot(BiQt, TK)

        self.tau[:] = cov.diagonal()
        self.tau -= dotd(QS[0], BiQtTK)
        self.tau[:] = 1/self.tau

        assert all(self.tau >= 0.)

        self.eta[:] = dot(cov, self._site.eta)
        self.eta[:] += self._mean
        self.eta[:] -= dot(QS[0], dot(BiQtTK, self._site.eta))
        self.eta[:] -= dot(QS[0], dot(BiQt, self._site.tau * self._mean))

        self.eta *= self.tau



import pytest

from numpy.testing import assert_allclose

from limix_inference.link import ProbitLink, LogitLink, LogLink
from limix_inference.link.link import Link

def test_probit_link():
    link = ProbitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    assert_allclose(link.latent_variance, 1.0)

def test_logit_link():
    link = LogitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    assert_allclose(link.latent_variance, 3.289868133696453)

def test_loglink_link():
    link = LogLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)
    with pytest.raises(NotImplementedError):
        print(link.latent_variance)

def test_link_interface():
    link = Link()

    with pytest.raises(NotImplementedError):
        print(link.value(None))

    with pytest.raises(NotImplementedError):
        print(link.inv(None))

    with pytest.raises(NotImplementedError):
        print(link.latent_variance)



from __future__ import absolute_import, division, unicode_literals

import logging

from numpy import asarray, dot, exp, zeros, log, ndarray, clip, concatenate
from numpy.linalg import LinAlgError

from numpy_sugar import epsilon

from liknorm import LikNormMachine
from optimix import Function, Scalar, Vector
from optimix.optimize import BadSolutionError

from ..ep import EP


class GLMM(EP, Function):
    r"""Expectation Propagation for Generalised Gaussian Processes.

    Let

    .. math::

        \mathrm Q \mathrm E \mathrm Q^{\intercal}
        = \mathrm G\mathrm G^{\intercal}

    be the eigen decomposition of the random effect's covariance.
    It turns out that the prior covariance of GLMM can be described as

    .. math::

        \mathrm Q s((1-\delta)\mathrm E
        + \delta\mathrm I) \mathrm Q^{\intercal}.

    This means that :math:`\mathrm Q` does not change during inference, and
    this fact is used here to speed-up EP inference for GLMM.

    Args:
        y (array_like): outcome variable.
        lik_name (str): likelihood name.
        mean (:class:`optimix.Function`): mean function.
                                          (Refer to :doc:`mean`.)
        cov (:class:`optimix.Function`): covariance function.
                                         (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy import dot
        >>> from numpy.random import RandomState
        >>>
        >>> from numpy_sugar.linalg import economic_qs
        >>>
        >>> from limix_inference.glmm import GLMM
        >>>
        >>> random = RandomState(0)
        >>> nsamples = 50
        >>>
        >>> X = random.randn(50, 2)
        >>> G = random.randn(50, 100)
        >>> K = dot(G, G.T)
        >>> ntrials = random.randint(1, 100, nsamples)
        >>> successes = [random.randint(0, i + 1) for i in ntrials]
        >>>
        >>> y = (successes, ntrials)
        >>>
        >>> QS = economic_qs(K)
        >>> glmm = GLMM(y, 'binomial', X, QS)
        >>> print('Before: %.4f' % glmm.feed().value())
        Before: -210.3568
        >>> glmm.feed().maximize(progress=False)
        >>> print('After: %.2f' % glmm.feed().value())
        After: -185.16
    """

    def __init__(self, y, lik_name, X, QS):
        super(GLMM, self).__init__()
        Function.__init__(
            self,
            beta=Vector(zeros(X.shape[1])),
            logscale=Scalar(0.0),
            logitdelta=Scalar(0.0))

        self._logger = logging.getLogger(__name__)

        if not isinstance(y, tuple):
            y = (y, )

        self._y = tuple([asarray(i, float) for i in y])
        self._X = X

        if not isinstance(QS, tuple):
            raise ValueError("QS must be a tuple.")

        if not isinstance(QS[0], tuple):
            raise ValueError("QS[0] must be a tuple.")

        Q = concatenate(QS[0], axis=1)
        S = zeros(Q.shape[0], dtype=float)
        S[:QS[1].shape[0]] = QS[1]

        self._QS = (Q, S)

        self._machine = LikNormMachine(lik_name, 500)
        self.set_nodata()

    def __del__(self):
        if hasattr(self, '_machine'):
            self._machine.finish()

    def _compute_moments(self):
        tau = self._cav['tau']
        eta = self._cav['eta']
        self._machine.moments(self._y, eta, tau, self._moments)

    def mean(self):
        return dot(self._X, self.beta)

    def fix(self, var_name):
        if var_name == 'scale':
            Function.fix(self, 'logscale')
        elif var_name == 'delta':
            Function.fix(self, 'logitdelta')
        elif var_name == 'beta':
            Function.fix(self, 'beta')
        else:
            raise ValueError("Unknown parameter name %s." % var_name)

    def unfix(self, var_name):
        if var_name == 'scale':
            Function.unfix(self, 'logscale')
        elif var_name == 'delta':
            Function.unfix(self, 'logitdelta')
        elif var_name == 'beta':
            Function.unfix(self, 'beta')
        else:
            raise ValueError("Unknown parameter name %s." % var_name)

    @property
    def scale(self):
        return exp(self.variables().get('logscale').value)

    @scale.setter
    def scale(self, v):
        self.variables().get('logscale').value = log(v)

    @property
    def delta(self):
        return 1 / (1 + exp(-self.variables().get('logitdelta').value))

    @delta.setter
    def delta(self, v):
        v = clip(v, epsilon.small, 1 - epsilon.small)
        self.variables().get('logitdelta').value = log(v / (1 - v))

    @property
    def beta(self):
        return self.variables().get('beta').value

    @beta.setter
    def beta(self, v):
        self.variables().get('beta').value = v

    def _eigval_derivative_over_logscale(self):
        x = self.variables().get('logscale').value
        d = self.delta
        E = self._QS[1]
        return exp(x) * ((1 - d) * E + d)

    def _eigval_derivative_over_logitdelta(self):
        x = self.variables().get('logitdelta').value
        s = self.scale
        c = (s * exp(x) / (1 + exp(-x))**2)
        E = self._QS[1]
        return c * (1 - E)

    def _S(self):
        s = self.scale
        d = self.delta
        return s * ((1 - d) * self._QS[1] + d)

    def value(self):
        r"""Log of the marginal likelihood.

        Args:
            mean (array_like): realised mean.
            cov (array_like): realised covariance.

        Returns:
            float: log of the marginal likelihood.
        """
        try:
            self._initialize(self.mean(), (self._QS[0], self._S()))
            self._params_update()
            lml = self._lml()
        except (ValueError, LinAlgError) as e:
            self._logger.info(str(e))
            raise BadSolutionError
        return lml

    def gradient(self):  # pylint: disable=W0221
        r"""Gradient of the log of the marginal likelihood.

        Args:
            mean (array_like): realised mean.
            cov (array_like): realised cov.
            gmean (array_like): realised mean derivative.
            gcov (array_like): realised covariance derivative.

        Returns:
            list: derivatives.
        """
        try:
            self._initialize(self.mean(), (self._QS[0], self._S()))
            self._params_update()

            dS0 = self._eigval_derivative_over_logitdelta()
            dS1 = self._eigval_derivative_over_logscale()

            grad = [self._lml_derivative_over_cov((self._QS[0], dS0)),
                    self._lml_derivative_over_cov((self._QS[0], dS1)),
                    self._lml_derivative_over_mean(self._X)]
        except (ValueError, LinAlgError) as e:
            self._logger.info(str(e))
            raise BadSolutionError

        return dict(logitdelta=grad[0],
                    logscale=grad[1],
                    beta=grad[2])



from numpy import add

from optimix import FunctionReduce


class SumMean(FunctionReduce):
    r"""Sum mean function.

    The mathematical representation is

    .. math::

        f(f_0, f_1, \dots) = f_0 + f_1 + \dots
    """
    def __init__(self, means):
        self._means = [c for c in means]
        FunctionReduce.__init__(self, self._means, 'sum')

    def value_reduce(self, values): # pylint: disable=R0201
        r"""Sum mean function evaluated at `(f_0, f_1, ...)`."""
        return add.reduce(list(values.values()))

    def gradient_reduce(self, _, gradients): # pylint: disable=R0201
        r"""Sum of mean function derivatives.

        Returns:
            :math:`f_0' + f_1' + \dots`
        """
        grad = dict()
        for (gn, gv) in iter(gradients.items()):
            for n, v in iter(gv.items()):
                grad[gn + '.' + n] = v
        return grad



from __future__ import division

from numpy import dot, log, maximum, sqrt, zeros
from numpy_sugar import epsilon
from numpy_sugar.linalg import ddot, economic_svd, solve

from .scan import FastScanner


class LMMCore(object):
    def __init__(self, y, X, QS):
        self._QS = QS
        self._y = y

        self._tM = None
        self.__tbeta = None

        self._svd = None
        self.X = X

    def get_fast_scanner(self):
        return FastScanner(self._y, self.X, self._QS, self.delta)

    @property
    def X(self):
        return dot(self._svd[0], ddot(self._svd[1], self._svd[2], left=True))

    @X.setter
    def X(self, X):
        self._svd = economic_svd(X)
        self._tM = ddot(self._svd[0], sqrt(self._svd[1]), left=False)
        self.__tbeta = zeros(self._tM.shape[1])

    @property
    def m(self):
        r"""Returns :math:`\mathbf m = \mathrm X \boldsymbol\beta`."""
        return dot(self._tM, self._tbeta)

    @property
    def delta(self):
        raise NotImplementedError

    @delta.setter
    def delta(self, _):
        raise NotImplementedError

    @property
    def _tbeta(self):
        return self.__tbeta

    @_tbeta.setter
    def _tbeta(self, value):
        self.__tbeta[:] = value

    @property
    def beta(self):
        return solve(self._svd[2].T, self._tbeta / sqrt(self._svd[1]))

    @beta.setter
    def beta(self, value):
        self._tbeta = sqrt(self._svd[1]) * dot(self._svd[2].T, value)

    @property
    def scale(self):
        a = [self._a(i) for i in [0, 1]]
        b = [self._b(i) for i in [0, 1]]
        c = [self._c(i) for i in [0, 1]]
        be = self.__tbeta
        p = [a[i] - 2 * b[i].dot(be) + be.dot(c[i]).dot(be) for i in [0, 1]]
        return maximum(sum(p) / len(self._y), epsilon.tiny)

    def _diag(self, i):
        if i == 0:
            return self._QS[1] * (1 - self.delta) + self.delta
        assert i == 1
        return self.delta

    def _a(self, i):
        return sum(self._yTQ_2x(i) / self._diag(i))

    def _b(self, i):
        return (self._yTQ(i) / self._diag(i)).dot(self._tMTQ(i).T)

    def _c(self, i):
        return (self._tMTQ(i) / self._diag(i)).dot(self._tMTQ(i).T)

    def _yTQ(self, i):
        return dot(self._y.T, self._QS[0][i])

    def _yTQ_2x(self, i):
        return self._yTQ(i)**2

    def _tMTQ(self, i):
        return self._tM.T.dot(self._QS[0][i])

    def _update_fixed_effects(self):
        nominator = self._b(1) - self._b(0)
        denominator = self._c(1) - self._c(0)
        self._tbeta = solve(denominator, nominator)

    def update(self):
        self._update_fixed_effects()

    def lml(self):
        self.update()

        n = len(self._y)
        p = n - self._QS[1].shape[0]
        LOG2PI = 1.837877066409345339081937709124758839607238769531250
        lml = -n * LOG2PI - n - n * log(self.scale)
        lml += -sum(log(self._diag(0))) - p * log(self._diag(1))
        lml /= 2
        return lml
